from dune.codegen.generation import (get_counted_variable,
                                     global_context,
                                     include_file,
                                     preamble,
                                     )
from dune.codegen.pdelab.driver import (FEM_name_mangling,
                                        get_trial_element,
                                        )
from dune.codegen.pdelab.driver.gridfunctionspace import (name_gfs,
                                                          name_leafview,
                                                          name_trial_gfs,
                                                          type_range,
                                                          type_trial_gfs,
                                                          preprocess_leaf_data,
                                                          )

from ufl.classes import Expr
from ufl import FiniteElement, MixedElement, TensorElement, VectorElement, TensorProductElement


def name_assembled_constraints():
    name = name_constraintscontainer()
    define_constraintscontainer(name)
    assemble_constraints(name)
    return name


def has_dirichlet_constraints(is_dirichlet):
    if isinstance(is_dirichlet, (tuple, list)):
        return any(has_dirichlet_constraints(d) for d in is_dirichlet)

    if isinstance(is_dirichlet, Expr):
        return True
    else:
        return bool(is_dirichlet)


@preamble(section="constraints")
def assemble_constraints(name):
    element = get_trial_element()
    gfs = name_trial_gfs()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    if has_dirichlet_constraints(is_dirichlet):
        bctype_function = name_bctype_function(element, is_dirichlet)
        return "Dune::PDELab::constraints({}, {}, {});".format(bctype_function,
                                                               gfs,
                                                               name,
                                                               )
    else:
        return "Dune::PDELab::constraints({}, {});".format(gfs,
                                                           name,
                                                           )


def name_bctype_function(element, is_dirichlet):
    # Note: Previously, there was a separate code branch for VectorElement here,
    #       which was implemented through PDELabs Power constraints concept.
    #       However this completely fails if you have different constraints for
    #       the children of a VectorElement. We therefore omit this branch completely.
    if isinstance(element, MixedElement):
        k = 0
        childs = []
        for subel in element.sub_elements():
            childs.append(name_bctype_function(subel, is_dirichlet[k:k + subel.value_size()]))
            k = k + subel.value_size()
        name = "{}_bctype".format("_".join(childs))
        define_composite_bctype_function(element, is_dirichlet, name, tuple(childs))
        return name
    else:
        assert isinstance(element, (FiniteElement, TensorProductElement))
        name = get_counted_variable("bctype")
        define_bctype_function(element, is_dirichlet[0], name)
        return name


@preamble(section="constraints")
def define_bctype_function(element, is_dirichlet, name):
    gv = name_leafview()
    bctype_lambda = name_bctype_lambda(name, is_dirichlet)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    return "auto {} = Dune::PDELab::makeBoundaryConditionFromCallable({}, {});".format(name,
                                                                                       gv,
                                                                                       bctype_lambda,
                                                                                       )


@preamble(section="constraints")
def define_composite_bctype_function(element, is_dirichlet, name, subgfs):
    include_file('dune/pdelab/constraints/common/constraintsparameters.hh', filetag='driver')
    return "Dune::PDELab::CompositeConstraintsParameters<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in subgfs),
                                                                             name,
                                                                             ', '.join(c for c in subgfs)
                                                                             )


def name_bctype_lambda(name, func):
    name = name + "_lambda"
    define_intersection_lambda(name, func)
    return name


@preamble(section="constraints")
def define_intersection_lambda(name, func):
    from ufl.classes import Expr
    if func is None:
        func = 0.
    if isinstance(func, (int, float)):
        return "auto {} = [&](const auto& x){{ return {}; }};".format(name, float(func))
    elif isinstance(func, Expr):
        from dune.codegen.pdelab.driver.visitor import ufl_to_code
        return "auto {} = [&](const auto& is, const auto& xl){{ {} }};".format(name, ufl_to_code(func))

    raise ValueError("Expression not understood")


@preamble(section="constraints")
def typedef_constraintscontainer(name):
    gfs = type_trial_gfs()
    r = type_range()
    return "using {} = {}::ConstraintsContainer<{}>::Type;".format(name, gfs, r)


def type_constraintscontainer():
    name = "{}_CC".format(type_trial_gfs())
    typedef_constraintscontainer(name)
    return name


@preamble(section="constraints")
def define_constraintscontainer(name):
    cctype = type_constraintscontainer()
    return ["{} {};".format(cctype, name), "{}.clear();".format(name)]


def name_constraintscontainer():
    gfs = name_trial_gfs()
    name = "{}_cc".format(gfs)
    element = get_trial_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    define_constraintscontainer(name)
    return name
